/*****

File: random.c
--------------
Implementation for the random.h interface

*****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h> 
#include <math.h>
#include <assert.h>

#include "random.h"
#include "nrutil.h"
#include "vector.h"
#include "randlib.h"
#include "statfcts.h"


/*#define RAND_MAX 30000 */


/* Universal constants 
#define  PI               3.141592653589793238  */

static short mother1[10];
static short mother2[10];
static short mStart=1;

#define m16Long 65536L				/* 2^16 */
#define m16Mask 0xFFFF          /* mask for lower 16 bits */
#define m15Mask 0x7FFF			/* mask for lower 15 bits */
#define m31Mask 0x7FFFFFFF     /* mask for 31 bits */
#define m32Double  4294967295.0  /* 2^32-1 */

/* Mother **************************************************************
|	George Marsaglia's The mother of all random number generators
|		producing uniformly distributed pseudo random 32 bit values with
|		period about 2^250.
|	The text of Marsaglia's posting is appended at the end of the function.
|
|	The arrays mother1 and mother2 store carry values in their
|		first element, and random 16 bit numbers in elements 1 to 8.
|		These random numbers are moved to elements 2 to 9 and a new
|		carry and number are generated and placed in elements 0 and 1.
|	The arrays mother1 and mother2 are filled with random 16 bit values
|		on first call of Mother by another generator.  mStart is the switch.
|
|	Returns:
|	A 32 bit random number is obtained by combining the output of the
|		two generators and returned in *pSeed.  It is also scaled by
|		2^32-1 and returned as a double between 0 and 1
|
|	SEED:
|	The inital value of *pSeed may be any long value
|
|	Bob Wheeler 8/8/94
*/


double Mother(unsigned long *pSeed)
{
	unsigned long   number, number1, number2;
	short           n, *p;
	unsigned short  sNumber;

		/* Initialize motheri with 9 random values the first time */
	if (mStart) {
		sNumber=*pSeed&m16Mask;   /* The low 16 bits */
		number=*pSeed&m31Mask;    /* Only want 31 bits */
		p=mother1;
		for (n=18;n--;) {
			number=30903*sNumber+(number>>16);   
                                          /* One line multiply-with-cary */
			*p++=sNumber=number&m16Mask;
			if (n==9)
				p=mother2;
		}
		                          /* make cary 15 bits */
		mother1[0]&=m15Mask;
		mother2[0]&=m15Mask;
		mStart=0;
	}

		                          /* Move elements 1 to 8 to 2 to 9 */
	memmove(mother1+2,mother1+1,8*sizeof(short));
	memmove(mother2+2,mother2+1,8*sizeof(short));
		                          /* Put the carry values in numberi */
	number1=mother1[0];
	number2=mother2[0];
		                          /* Form the linear combinations */
	number1+=1941*mother1[2]+1860*mother1[3]+1812*mother1[4]+1776*mother1[5]+
				1492*mother1[6]+1215*mother1[7]+1066*mother1[8]+12013*mother1[9];
	number2+=1111*mother2[2]+2222*mother2[3]+3333*mother2[4]+4444*mother2[5]+
				5555*mother2[6]+6666*mother2[7]+7777*mother2[8]+9272*mother2[9];
                         		  /* Save the high bits of numberi as the new carry */
	mother1[0]=number1/m16Long;
	mother2[0]=number2/m16Long;
		                          /* Put the low bits of numberi into motheri[1] */
	mother1[1]=m16Mask&number1;
	mother2[1]=m16Mask&number2;
		                          /* Combine the two 16 bit random numbers into one 32 bit */
	*pSeed=(((long)mother1[1])<<16)+(long)mother2[1];
		                          /* Return a double value between 0 and 1 */
	return ((double)*pSeed)/m32Double;
}

int gengeo(double p)
// return geometric variate (number of trials until first success)
{
  int count = 1;

  while (genunf(0, 1) > p)
    count++;

  return(count);
}

VectorDouble runif(int n, double low, double high)
// put n uniforms in vector
{
  VectorDouble x(n);

  //unsigned long Seed, *pSeed;
  //Seed = (unsigned long) time(NULL);
  //pSeed = &Seed;

  for (int i = 0; i < n; i++)
    x[i] = genunf(low, high);

  return(x);
}

VectorDouble rtriangle(int n, double low, double high)
// put n sums of two uniforms in vector
{
  VectorDouble x(n);

  for (int i = 0; i < n; i++)
    x[i] = (genunf(low, high) + genunf(low, high))/2.0;

  return(x);
}

VectorDouble rtriangleInv(int n, double low, double high)
// put n "inverted" sums of two uniforms in vector
{
  double mean = (low + high)/2.0, three_mean = 1.5 * (low + high);
  VectorDouble x(n);

  for (int i = 0; i < n; i++) {
    if (genunf(0, 1) < 0.5)
      x[i] = mean - (genunf(low, high) + genunf(low, high))/2.0;
    else
      x[i] = three_mean - (genunf(low, high) + genunf(low, high))/2.0;
  }

  return(x);
}

VectorDouble rtriangleSkew(int n, double low, double high)
// put n varites with density "proportional to x" in vector
{
  double dist_square = pow(high - low, 2);
  VectorDouble x(n);

  for (int i = 0; i < n; i++)
    x[i] = low + sqrt(genunf(low, high) * dist_square);

  return(x);
}


int MotherInt(int low, int high, unsigned long *pSeed)
     /* Function: MotherInt
	-------------------
	Returns a random integer in the range low to high, inclusive;
	uses Mother to produce a random uniform variate as starting point
     */

{
  int k;
  double p;

  p = Mother(pSeed);
  k = (int) (p * (high - low + 1));
  return (low + k);
}

VectorInt rint(int n, double low, double high)
// put n random integers in vector
{
  VectorInt Int(n);

  //unsigned long Seed, *pSeed;
  //Seed = (unsigned long) time(NULL);
  //pSeed = &Seed;

  for (int i = 0; i < n; i++)
    Int[i] = ignuin(low, high);

  return(Int);
}

VectorDouble rlogist(int n, double mu, double sigma)
// put n logistics with mean mu and variance sigma in vector
{
  double b;
  const double pi = 3.141593;
  VectorDouble x(n);

  b = sqrt(3 * sigma) / pi;

  for (int i = 0; i < n; i++)
    x[i] = genlogist(mu, b);

  return(x);
}

double gasdev(unsigned long *pSeed)
// returns a N(0,1) random variable
{
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*Mother(pSeed)-1.0;
			v2=2.0*Mother(pSeed)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}

VectorDouble rnorm(int n, double mu, double sigma)
// put n normals in vector
{
  VectorDouble x(n);

  //unsigned long Seed, *pSeed;
  //Seed = (unsigned long) time(NULL);
  //pSeed = &Seed;

  for (int i = 0; i < n; i++)
    x[i] = gennor(mu, sigma);

  return(x);
}

VectorDouble rnorm(int n)
// put n normals in vector
{
  double mu = 0, sigma = 1;
  VectorDouble x(n);

  //unsigned long Seed, *pSeed;
  //Seed = (unsigned long) time(NULL);
  //pSeed = &Seed;

  for (int i = 0; i < n; i++)
    x[i] = gennor(mu, sigma);

  return(x);
}

VectorDouble rnormMult(int n)
// put n normal-products in vector
// uncorrelated but k-dependent (k picked below ...)
{
  int k = 2;
  double mu = 0, sigma = 1;
  VectorDouble x = rnorm(n + k, mu, sigma);

  for (int i = 0; i < n; i++)
    x[i] = prod(x(i, i + k));

  return(x(0, n - 1));
}

VectorDouble rbeta(int n, double a, double b)
// n iid Beta(a,b) variables
{
  VectorDouble x(n);

  for (int i = 0; i < n; i++)
    x[i] = genbet(a, b);

  return(x);
}

VectorDouble rpareto(int n, double a)
// n iid Pareto(a) variables
{
  VectorDouble x;

  x = runif(n);

  for (int i = 0; i < n; i++)
    x[i] = pow((1/(1 - x[i])), 1.0/a);

  return(x);
}

VectorDouble rparetoMean0(int n, double a)
// n iid centered Pareto(a) variables
{
  VectorDouble x(n);

  assert(a > 1);

  for (int i = 0; i < n; i++)
    x[i] = pow(1.0/genunf(0, 1), 1.0/a) - (a /(a - 1.0));

  return(x);
}

VectorDouble rparetoSym(int n, double a)
// n iid centered Pareto(a) variables
{
  VectorDouble x(n);

  for (int i = 0; i < n; i++) {
    if (genunf(0, 1) < 0.5)
      x[i] = pow(1.0/genunf(0, 1), 1.0/a) - 1;
    else
      x[i] = 1 - pow(1.0/genunf(0, 1), 1.0/a);
  }

  return(x);
}


VectorDouble rt3(int n)
// t(3) dist       ??????????
{
  VectorDouble x(n);

  for (int i = 0; i < n; i++)
    if (genunf(0, 1) < 0.5)
      x[i] = sqrt(genf(1, 3));
    else
      x[i] = - sqrt(genf(1, 3));

  return(x);
}

VectorDouble rtwo(int n, double x1, double x2, double p1)
// n iid variables from two point distribution
{
  VectorDouble x(n), unif;

  unif = runif(n);
  for (int i = 0; i < n; i++) {
    if (unif[i] < p1)
      x[i] = x1;
    else
      x[i] = x2;
  }
  return(x);
}

VectorDouble rtwoA(int n, double x1, double x2)
// n iid variables from two point distribution
{
  VectorDouble x(n);
  double p1 = 0.95;

  for (int i = 0; i < n; i++) {
    if (genunf(0, 1) < p1)
      x[i] = x1;
    else
      x[i] = x2;
  }
  return x;
}


void Sample_with(double *x, int n, double *x_sample, int b, unsigned long *pSeed)

     /* Function: Sample_with
	---------------------
        Samples b data points with replacement from the vector
        x[0..n-1] and stores them in x_sample[0..b-1]
     */
{
  int i, index;

  for (i = 0; i < b; i++) {
    index = MotherInt(0, n-1-i, pSeed);
    x_sample[i] = x[index];
  }
}


void Sample_without(double *x, int n, double *x_sample, int b,
                    unsigned long *pSeed)

     /* Function: Sample_without
	------------------------
        Samples b data points without replacement from the vector
        x[0..n-1] and stores them in x_sample[0..b-1]
     */
     
{
  int i, index;
  double *x_copy;

  x_copy = dvector(0, n-1);
  for (i = 0; i < n; i++)
    x_copy[i] = x[i];

  for (i = 0; i < b; i++) {
    index = MotherInt(0, n-1-i, pSeed);
    x_sample[i] = x_copy[index];
    x_copy[index] = x_copy[n-1-i];
  }

  free_dvector(x_copy, 0, n-1);
}

VectorInt sample(const VectorInt &x, int b, int with)
// samples with our without replacement from an int vector
{
  int i, n, index;
  VectorInt randomInt, xSample(b);

  if (with) {
    randomInt = rint(b, 0, x.getSize() - 1);
    for (i = 0; i < b; i++)
      xSample[i] = x[randomInt[i]];
    return(xSample);
  }
  else {
    n = x.getSize();
    assert(b <= n);
    VectorInt xCopy = x;
    for (i = 0; i < b; i++) {
      index = ignuin(0, n - 1 - i);
      xSample[i] = xCopy[index];
      xCopy[index] = xCopy[n - 1 - i];
    }
    return(xSample);
  }
}

VectorDouble sample(const VectorDouble &x, int b, int with)
// samples with our without replacement from a double vector
{
  int i, n, index;
  VectorDouble xSample(b);
  VectorInt randomInt;

  if (with) {
    randomInt = rint(b, 0, x.getSize() - 1);
    for (i = 0; i < b; i++)
      xSample[i] = x[randomInt[i]];
    return(xSample);
  }
  else {
    n = x.getSize();
    assert(b <= n);
    VectorDouble xCopy = x;
    for (i = 0; i < b; i++) {
      index = ignuin(0, n - 1 - i);
      xSample[i] = xCopy[index];
      xCopy[index] = xCopy[n - 1 - i];
    }
    return(xSample);
  }
}


