#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <assert.h>

#include "time_sim.h"
#include "statfcts.h"
#include "nrutil.h"
#include "random.h"
#include "tisu.h"
#include "vector.h"
#include "matrix.h"
#include "math.h"
#include "randlib.h"
#include "boot.h"


VectorDouble expSim(int n, int nStart)
// generates dependent exp(1) process according to Kitamura (1999)
{
  int i;
  VectorDouble norm1 = rnorm(n + nStart), norm2 = rnorm(n + nStart),
               x(n + nStart);

  for (i = 1; i < n + nStart; i++) {
    x[i] = (pow(norm1[i] + 0.7 * norm2[i - 1], 2) +
           pow(norm2[i] + 0.7 * norm2[i - 1], 2)) / 2.98 - 1.0;
  }

  return x(nStart, nStart + n - 1);
}


VectorDouble Markov2Sim(int n, VectorDouble piVec, int mean0, int nStart)
/*
  - generates 2 state Markov chain with values 0 and 1
  - piVec[0] = pi_00 and piVec[1] = pi_11
  - if mean0, then mean of stationary distribution is substracted
*/
{
  int i;
  double eps = 0.01, mean;
  VectorDouble x(n + nStart);

  for (i = 1; i < n + nStart; i++) {
    if (x[i - 1] < eps) {
      if (genunf(0, 1) < piVec[0])
        x[i] = 0;
      else
        x[i] = 1;
    }
    else {
      if (genunf(0, 1) < piVec[1])
        x[i] = 1;
      else
        x[i] = 0;
    }
  }

  if (mean0) {
    mean = (1 - ((1 - piVec[1]) / (2 - sum(piVec))));
    return (x - mean)(nStart, n + nStart - 1);
  }
  else
    return x(nStart, n + nStart - 1);
}


VectorDouble rgarch11(int n, double omega, double alpha, double beta,
                      int n_start)
// generates a conditionally normal GARCH(1,1) sequence with mean 0
{
  int     i;
  double  h;
  VectorDouble x(n_start + n), epsilon(n_start + n);

  epsilon = rnorm(n_start + n, 0, 1);

/* initialize the h and epsilon sequences with something meaningful */

  if (alpha + beta < 0.999) {
    h = omega / (1 - alpha - beta);
    x[0] = epsilon[0] * sqrt(h);
  }
  else {
    h = 0;
    x[0] = 0;
  }

/* compute the GARCH sequence recursively */

  for (i = 1; i < n_start + n; i++) {
    h = omega + alpha * pow(x[i - 1], 2) + beta * h;
    x[i] = epsilon[i] * sqrt(h);
  }

  return x(n_start, n_start + n - 1);
}


VectorDouble armaSim(int n, double mu, const VectorDouble &ar,
                     const VectorDouble &ma, VectorDouble (*randgen)(int m),
                     int nStart)
{
  int i, j, k, n_ar = ar.getSize(), n_ma = ma.getSize(), first = max(n_ar, n_ma);
  VectorDouble innovations = (*randgen) (n + nStart), x(innovations);

  for (i = first; i < n + nStart; i++) {
    for (j = 0; j < n_ar; j++)
      x[i] += ar[j] * x[i - j - 1];
    for (k = 0; k < n_ma; k++)
      x[i] += ma[k] * innovations[i - k - 1];
    x[i] += mu;
  }

  return x(nStart, nStart + n - 1);
}


VectorDouble armaSim(int n, double mu, const VectorDouble &ar,
                     const VectorDouble &ma, const VectorDouble &innovations,
                     int nStart)
/*
- specify vector of innovations rather than generate it inside of the function
- for example, used for stable innovations where no C++ code is available
*/
{
  int i, j, k, n_ar = ar.getSize(), n_ma = ma.getSize(), first = max(n_ar, n_ma);
  VectorDouble x(innovations);

  assert(innovations.getSize() >= n + nStart);

  for (i = first; i < n + nStart; i++) {
    for (j = 0; j < n_ar; j++)
      x[i] += ar[j] * x[i - j - 1];
    for (k = 0; k < n_ma; k++)
      x[i] += ma[k] * innovations[i - k - 1];
    x[i] += mu;
  }

  return x(nStart, nStart + n - 1);
}



void ar_trend_sim(int n, double mu, double beta, double *ar,
                  int n_ar, double *innov, int n_start)
// generate an AR(p) sequence with time trend
// mu is mean parameter and beta is time trend paramter
// innov contains innovations and will "export" the series
{
  int     i, j;
  double  *time_series;

  time_series = dvector(0, n_start + n - 1);

  for (i = 0; i < n_ar; i++){
    time_series[i] = 0;
  }
  for (i = n_ar; i < n_start + n; i++) {
    time_series[i] = mu + (i - n_start) * beta;
    for (j = 0; j < n_ar; j++) {
      time_series[i] += time_series[i-1-j] * ar[j];
    }
    time_series[i] += innov[i];
  }
  for (i = 0; i < n; i++){
    innov[i] = time_series[i+ n_start];
  }
  free_dvector(time_series, 0, n + n_start - 1);
}

VectorDouble arTrendSim(int n, double mu, double beta, const VectorDouble &ar,
                        VectorDouble (*randgen)(int m), int n_start)
// generate an AR(p) sequence with time trend
// mu is mean parameter and beta is time trend paramter
{
  int     i, j, n_ar = ar.getSize();
  VectorDouble time_series(n_start + n), innov = (*randgen)(n + n_start);

  for (i = 0; i < n_ar; i++){
    time_series[i] = 0;
  }
  for (i = n_ar; i < n_start + n; i++) {
    time_series[i] = mu + (i - n_start) * beta;
    for (j = 0; j < n_ar; j++) {
      time_series[i] += time_series[i-1-j] * ar[j];
    }
    time_series[i] += innov[i];
  }
  return time_series(n_start, n_start + n - 1);
}

VectorDouble arTrendBootSim(int n, double mu, double beta, const VectorDouble &ar,
                            VectorDouble fitted_innov, int b,
                            VectorDouble (*boot) (const VectorDouble &, int, int, int),
                            int n_start)
// generate an AR(p) sequence with time trend
// mu is mean parameter and beta is time trend paramter
// innovations are bootstrapped from Vector of fitted innovations
{
  int     i, j, n_ar = ar.getSize();
  VectorDouble time_series(n_start + n),
               innov = (*boot)(fitted_innov, b, 0, n + n_start);

  for (i = 0; i < n_ar; i++){
    time_series[i] = 0;
  }
  for (i = n_ar; i < n_start + n; i++) {
    time_series[i] = mu + (i - n_start) * beta;
    for (j = 0; j < n_ar; j++) {
      time_series[i] += time_series[i-1-j] * ar[j];
    }
    time_series[i] += innov[i];
  }
  return time_series(n_start, n_start + n - 1);
}

VectorDouble ArmaTrendSim(int n, double mu, double beta, const VectorDouble &ar,
                          double ma, VectorDouble (*randgen)(int m), int n_start)
// generate an AR(p) sequence with time trend
// mu is mean parameter and beta is time trend paramter
// allow MA(1) innovation process
{
  int     i, j, n_ar = ar.getSize();
  VectorDouble time_series(n_start + n), innov = (*randgen)(n + n_start);

  for (i = 0; i < n_ar; i++){
    time_series[i] = 0;
  }
  for (i = n_ar; i < n_start + n; i++) {
    time_series[i] = mu + (i - n_start) * beta;
    for (j = 0; j < n_ar; j++) {
      time_series[i] += time_series[i-1-j] * ar[j];
    }
    time_series[i] += innov[i] + ma * innov[i - 1];
  }
  return time_series(n_start, n_start + n - 1);
}


MatrixDouble VAR1SimNorm(int n, const MatrixDouble &Var, const VectorDouble &start)
// generates VAR(1) data using i.i.d. standard normal innovations
// if Var matrix is p x p, no intercept is included; otherwise it is (p + 1) x p
{
  assert(Var.getRow() <= Var.getCol() && Var.getCol() <= Var.getRow() + 1);
  assert(Var.getRow() == start.getSize());

  int i, j, p = Var.getRow(), addInt = 0;
  MatrixDouble x(n, p);

  if (Var.getCol() == Var.getRow() + 1) addInt = 1;

  // initialize
  assignRow(x, 0, start);

  // compute remaining obervations recursively
  for (i = 1; i < n; i++)
    for (j = 0; j < p; j++) {
      if (addInt)
        x(i, j) = innerProduct(c(1, extractRow(x, i - 1)), extractRow(Var, j))
                  + gennor(0.0, 1.0);
      else
        x(i, j) = innerProduct(extractRow(x, i - 1), extractRow(Var, j))
                  + gennor(0.0, 1.0);
    }

  return x;
}

MatrixDouble VAR1SimBoot(int n, const MatrixDouble &Var, const VectorDouble &start,
                         const MatrixDouble &res, int b, int statBoo)
// generates VAR(1) data using bootstrapped residuals
// if Var matrix is p x p, no intercept is included; otherwise it is (p + 1) x p
{
  assert(Var.getRow() <= Var.getCol() && Var.getCol() <= Var.getRow() + 1);
  assert(Var.getRow() == res.getCol());
  assert(Var.getRow() == start.getSize());

  int i, j, p = Var.getRow(), addInt = 0;
  MatrixDouble x(n, p), resStar(n - 1, p);

  if (b == 1)
    resStar = bootstrapMB(res, 1, 0, n - 1);
  else {
    if (statBoo)
      resStar = bootstrapStat(res, b, 0, n - 1);
    else
      resStar = bootstrapMBcircle(res, b, 0, n - 1);
  }

  if (Var.getCol() == Var.getRow() + 1) addInt = 1;

  // initialize
  assignRow(x, 0, start);

  // compute remaining obervations recursively
  for (i = 1; i < n; i++)
    for (j = 0; j < p; j++) {
      if (addInt)
        x(i, j) = innerProduct(c(1.0, extractRow(x, i - 1)), extractRow(Var, j))
                  + resStar(i - 1, j);
      else
        x(i, j) = innerProduct(extractRow(x, i - 1), extractRow(Var, j))
                  + resStar(i - 1, j);
    }

  return x;
}


MatrixDouble VAR1_LM(int n, double rho, double theta, double lambda, int n_start)
// generates VAR(1) errors of Li + Maddala (1997) simulations
// first column corresponds to x data and second column to y data
{
  assert(0 <= lambda && lambda < 1);
  assert(0 <= rho && rho < 1);
  assert(0 <= theta && theta < 1);

  int i;
  double n1, n2, factor = pow(1 - pow(lambda, 2), 0.5);
  MatrixDouble x = nullMatrixDouble(n + n_start, 2);

  // compute remaining obervations recursively
  for (i = 1; i < n + n_start; i++) {
    n1 = gennor(0, 1);
    n2 = gennor(0, 1);
    n2 = lambda * n1 + factor * n2;
    x(i, 0) = theta * x(i - 1, 0) + n1;
    x(i, 1) = rho * x(i - 1, 1) + n2;
  }

  cout << endl << var(extractRows(x, n_start, n_start + n - 1));
  return extractRows(x, n_start, n_start + n - 1);
}


VectorDouble GarchSimBoot(int n, double omega, double alpha, double beta,
                          const VectorDouble &res, int nStart)
/*
- generates GARCH(1,1) sequence according to given parameter values
- innovations are Efron-bootstrapped from a given vector res
*/
{
  int i;
  double h;
  VectorDouble innov, x(n + nStart);

  innov = bootstrap(res, 0, n + nStart);

  if (alpha + beta < 0.99999)
    h = omega / (1.0 - alpha - beta);
  else
    h = 0;

  x[0] = sqrt(h) * innov[0];
  for (i = 1; i < n + nStart; i++) {
    h = omega + alpha * pow(x[i - 1], 2) + beta * h;
    x[i] = sqrt(h) * innov[i];
  }

  return x(nStart, nStart + n - 1);
}


MatrixDouble DGP_LM(int n, double beta, double rho, double theta, double lambda,
                    int n_start)
// DGP of Li + Maddala (1997) simulations
// first column corresponds to x data and second column to y data
{
  int i;
  MatrixDouble errors, data(n, 2);

  errors = VAR1_LM(n, rho, theta, lambda, n_start);
  data(0, 0) = errors(0, 0);
  data(0, 1) = errors(0, 1);
  for (i = 1; i < n; i++) {
    data(i, 0) += data(i - 1, 0) + errors(i, 0);
    data(i, 1) += beta * data(i, 0) + errors(i, 1);
  }

  return data;
}


MatrixDouble DGP_H99(int n, double d2, int zISq, VectorDouble (*randgen)(int m))
/*
- DGP of Hansen (1999)
- first two columns of output is x matrix, then comes y vector, then q vector
*/
{
  int i;
  VectorDouble z, y(n), q, e;
  MatrixDouble help1, help2;

  q = (*randgen)(n) + 2;
  e = (*randgen)(n);
  if (zISq)
    z = q;
  else
    z = (*randgen)(n);
  for (i = 0; i < n; i++) {
    if (q[i] <= 2)
      y[i] = 1 + d2 * z[i] + e[i];
    else
      y[i] = e[i];
  }
  help1 = cbind(1, z);
  help2 = cbind(y, q);

  return cbind(help1, help2);
}

VectorDouble DGP_CT98(int n, double p0, double pa, double pb, double gamma,
                      double s2Mult, int nStart)
/*
- implements DGP of Chan and Tsay (1998)
- p0 = phi_0, pa = phi_1-, pb = phi_1+, gamma = r
*/
{
  int i;
  VectorDouble x(n + nStart);

  x[0] = 1;
  for (i = 1; i < n + nStart; i++) {
    if (x[i - 1] <= gamma)
      x[i] = p0 + pa * (x[i - 1] - gamma) + gennor(0, 1);
    else
      x[i] = p0 + pb * (x[i - 1] - gamma) + s2Mult * gennor(0, 1);
  }

  return x(nStart, n + nStart - 1);
}


VectorDouble DGP_TARdc(int n, const VectorDouble &theta, double s2Mult,
                       int nStart)
/*
- implements DGP of discontinuous TAR such as in Chan (1993)
- of course, can be made continuous by proper parameter choice
- theta = c(gamma, beta1, beta2, alpha1, alpha2)
- beta's are parameters when x(t - 1) <= gamma
- innovations are i.d. normal with possibly different variance
*/
{
  int i;
  VectorDouble x(n + nStart);

  x[0] = 1;
  for (i = 1; i < n + nStart; i++) {
    if (x[i - 1] <= theta[0])
      x[i] = theta[1] + theta[2] * x[i - 1] + gennor(0, 1);
    else
      x[i] = theta[3] + theta[4] * x[i - 1] + s2Mult * gennor(0, 1);
  }

  return x(nStart, n + nStart - 1);
}


VectorDouble DGP_TARdcBoot(int n, const VectorDouble &theta,
                           const VectorDouble &resid1, const VectorDouble &resid2,
                           int sameResid, double x0, int nStart)
/*
- implements DGP of discontinuous TAR such as in Chan (1993)
- of course, can be made continuous by proper parameter choice
- theta = c(gamma, beta1, beta2, alpha1, alpha2)
- beta's are parameters when x(t - 1) <= gamma
- innovations are resampled from fitted residuals, either from common vector
  resid1 or, when different variances are suspected, from vectors resid1
  and resid2
*/
{
  int i;
  VectorDouble x(n + nStart);

  x[0] = x0;
  if (sameResid) {
    VectorDouble epsStar = sample(resid1, n + nStart, 1);
    for (i = 1; i < n + nStart; i++) {
      if (x[i - 1] <= theta[0])
        x[i] = theta[1] + theta[2] * x[i - 1] + epsStar[i];
      else
        x[i] = theta[3] + theta[4] * x[i - 1] + epsStar[i];
    }
  }
  else {
    for (i = 1; i < n + nStart; i++) {
      if (x[i - 1] <= theta[0])
        x[i] = theta[1] + theta[2] * x[i - 1] + sample(resid1, 1, 1)[0];
      else
        x[i] = theta[3] + theta[4] * x[i - 1] + sample(resid2, 1, 1)[0];
    }
  }

  return x(nStart, n + nStart - 1);
}


MatrixDouble DGP_BR_Horo(int n, int uDist, const VectorDouble &beta,
                         double uMult)
/*
- DGP for binary response model
- uses DGP of Horowitz (1992) with Manski's normalization
- uDist specifies one of the four distributions for the innovations u
*/
{
  int i;
  double z;
  VectorDouble u(n);
  MatrixDouble data(n, 3);

  assignCol(data, 0, rnorm(n, 0, 1));
  assignCol(data, 1, rnorm(n, 0, 1));
  if (uDist == 1)
    u = rlogist(n, 0, 1);
  else if (uDist == 2)
    u = sqrt(12.0) * (runif(n) - 0.5);
  else if (uDist == 3)
    u = sqrt(1.0/3.0) * rt3(n);
  else if (uDist == 4) {
    for (i = 0; i < n; i++) {
      z = data(i, 0) + data(i, 1);
      u[i] = 0.25 * (1 + 2 * pow(z, 2) + pow(z, 4)) * genlogist(0, 1);
    }
  }
  else
    cout << "invalid choice of uDist";

  for (i = 0; i < n; i++)
    if (beta[0] * data(i, 0) + beta[1] * data(i, 1) + uMult * u[i] >= 0)
      data(i, 2) = 1.0;
    else
      data(i, 2) = 0.0;

  return data;
}


MatrixDouble AR_Homo(int n, int p, double rho, VectorDouble (*randgen)(int m),
                     int nStart)
/*
- generates regression data using the AR(1)-HOMO model of Andrews (1991)
- scales to achieve unconditional variance indpendent of rho
- p specifies the number of regressors, including the intercept
- first k columns of the output are the regressors, last column is y
- true beta is equal to zero without loss of generality
*/
{
  int k;
  MatrixDouble result(n, p + 1);

  assignCol(result, 0, rep(1.0, n));
  for (k = 1; k < p + 1; k++)
    assignCol(result, k, armaSim(n, 0, vec(rho), vec(0.0), randgen, nStart) * sqrt(1 - pow(rho, 2)));

  return result;
}


MatrixDouble AR_Het1(int n, int p, double rho, VectorDouble (*randgen)(int m),
                     int nStart)
/*
- generates regression data using the AR(1)-HET1 model of Andrews (1991)
- scales to achieve unconditional variance indpendent of rho
- p specifies the number of regressors, including the intercept
- first k columns of the output are the regressors, last column is y
- true beta is equal to zero without loss of generality
*/
{
  int i, k;
  MatrixDouble result(n, p + 1);

  assignCol(result, 0, rep(1.0, n));
  for (k = 1; k < p + 1; k++)
    assignCol(result, k, armaSim(n, 0, vec(rho), vec(0.0), randgen, nStart) * sqrt(1 - pow(rho, 2)));
  for (i = 0; i < n; i++)
    result(i, p) *= fabs(result(i, 1)); 

  return result;
}


MatrixDouble MA_Homo(int n, int p, double theta, VectorDouble (*randgen)(int m),
                     int nStart)
/*
- generates regression data using the MA(1)-HOMO model of Andrews (1991)
- scales to achieve unconditional variance indpendent of rho
- p specifies the number of regressors, including the intercept
- first k columns of the output are the regressors, last column is y
- true beta is equal to zero without loss of generality
*/
{
  int k;
  MatrixDouble result(n, p + 1);

  assignCol(result, 0, rep(1.0, n));
  for (k = 1; k < p + 1; k++)
    assignCol(result, k, armaSim(n, 0, vec(0.0), vec(theta), randgen, nStart) / sqrt(1 + pow(theta, 2)));

  return result;
}
