/*
   In this 'package' different methods for making inference about
   parameters of dependent data (time series) are implemented.
   The methods are:
      - "subsampling" (the subsampling method by Politis/Romano)
      - "moving_blocks" (Kuensch)
      - "stationary_bootstrap" (Politis/Romano)
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <string.h>

#include "sort.h"
#include "statfcts.h"
#include "random.h"
#include "nrutil.h"
#include "tisu.h"


#define D_STYLE "%10.6f"
#define pi 3.141593



double emp_quantile(double p, double *x, int x_length)
/*
   - computes p-th quantile of a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - uses formula: "p * (x_lenght - 1) + 1rst" smallest value
   - also uses interpolation
*/
{
  double index, lambda;
  int    index_floor;

  index = p * (x_length - 1.0);
  index_floor = index/1;
  lambda = index - index_floor;
  return((1.0 - lambda) * x[index_floor] +  lambda * x[index_floor + 1]);

}

double emp_quantile(double p, const VectorDouble &x)
/*
   - computes p-th quantile of a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - uses formula: "p * (x_length - 1) + 1rst" smallest value
   - also uses interpolation
*/
{
  double index, lambda, x_length;
  int    index_floor;

  x_length = x.getSize();
  index = p * (x_length - 1.0);
  index_floor = index/1;
  lambda = index - index_floor;
  return((1.0 - lambda) * x[index_floor] +  lambda * x[index_floor + 1]);
}

double emp_combined_quantile(double p, double *points, double *value, int n_comb)
/*
   - computes p-th quantile of a discrete `combined cdf',
     given by "points" and "values"
   - the two vectors  must be in ascending order!
   - finds `conservative' quantiles for construction of CI's
*/

{
  int  index;

  if (p < 0.5) {
    index = 0;
    while (value[index] <= p) index++;
    index -= 1;
  }

  else {
    index = n_comb - 1;
    while (value[index] >= p) index--;
    index += 1;
  }

  return(points[index]);
  
}

double upper_Pvalue(double observed, double *x, int x_length)
/*
   - computes upper P-value from "observed" value and a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - P-value is defined as the proportion of the x's greater than or
     equal to the observed value
*/

{
  int count;
  if (x[0] >= observed) return (1.0);
  else {
     count = 0;
     while (x[count] < observed && count < x_length) count++;
     return (1.0 - (count + 0.0)/(1.0 * x_length));
   }
}

double lower_Pvalue(double observed, double *x, int x_length)
/*
   - computes lower P-value from "observed" value and a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - P-value is defined as the proportion of the x's less than or
     equal to the observed value
*/

{
  int count;

  if (x[x_length - 1] <= observed) return (1.0);
  else if (x[0] > observed) return (0.0);
  else {
     count = x_length - 1;
     while (x[count] >= observed && count >= 0) count--;
     return ((count + 1.0)/(1.0 * x_length));
   }
}

double upper_Pvalue(double observed, const VectorDouble &x)
/*
   - computes upper P-value from "observed" value and a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - P-value is defined as the proportion of the x's greater than or
     equal to the observed value
*/

{
  int count, x_length = x.getSize();
  if (x[0] >= observed) return (1.0);
  else if (x[x_length - 1] < observed) return (0.0);
  else {
     count = 0;
     while (x[count] < observed && count < x_length) count++;
     return (1.0 - (count + 0.0)/(1.0 * x_length));
   }
}

double lower_Pvalue(double observed, const VectorDouble &x)
/*
   - computes lower P-value from "observed" value and a discrete
     reference distribution, given by "x"
   - the vector "x" must be in ascending order!
   - P-value is defined as the proportion of the x's less than or
     equal to the observed value
*/

{
  int count, x_length = x.getSize();

  if (x[x_length - 1] <= observed) return (1.0);
  else if (x[0] > observed) return (0.0);
  else {
     count = x_length - 1;
     while (x[count] >= observed && count >= 0) count--;
     return ((count + 1.0)/(1.0 * x_length));
   }
}


int length_sub(int n, int b, int lag)
{
  if (lag >= b){
     return (n + (lag - b))/lag;
     }
  else{
     return (n - (b - lag))/(b -(b - lag));
   }
}

void display_ivector(int *v, int nl, int nh)
/* display an int vector with subscript range v[nl..nh] */
{
        int i;

	for(i=nl;i<=nh;i++) printf("%d ", v[i]);
        printf("\n");
}

void display_dvector(double *v, int nl, int nh)
/* display a double vector with subscript range v[nl..nh] */
{
        int i;
       
	for(i=nl;i<=nh;i++) printf(D_STYLE, v[i]);
        printf("\n");
}

void display_dmatrix(double **m, long nrl, long nrh, long ncl, long nch)
/* display a double matrix with subscript range m[nrl..nrh][ncl..nch] */ 
{
        long i,j;

        for(i=nrl;i<=nrh;i++) {
		for(j=ncl;j<=nch;j++)
                        printf(D_STYLE, m[i][j]);
		printf("\n");
        }
        printf("\n");
}



